#!/usr/bin/env bash
# Author: Zhang Huangbin <zhb@iredmail.org>
# Purpose: Install Spider Email Archiver on systemd-based Linux distributions,
#          includes but not limited to:
#           - CentOS, Rocky Linux, AlmaLinux
#           - Debian, Ubuntu
#
# Sample usage:
#
#   bash install.sh

export LC_ALL=C
export PROG="Spider Email Archiver"

# Daemon user/group.
export USER="spider"
export GROUP="spider"

# Use fixed uid/gid for easier migration.
export USER_ID="21000"
export GROUP_ID="21000"

############################################
# Directory tree
############################################
#
# /opt/spider/
#       |- settings.json    # Main config file.
#       |- archive/         # Archived email messages and attachments.
#       |- db/              # Databases.
#       |- fts/email/       # Full-text indexed data.
#       |- log/             # Log files.
#       |- cert.pem         # SSL cert file (or symbol link).
#       |- key.pem          # SSL private key file (or symbol link).
#       |- secret.key       # Secret key used to encrypt / decrypt email
#                           # messages and attachments.
#
# Directory used to store application and data.
export DATA_DIR="/opt/spider"
export CONF="${DATA_DIR}/settings.json"

# Directory used to store archived email messages and attachments.
# WARNING: All archived data will be saved under this directory, please make
#          sure it has enough disk space.
export STORAGE_DIR="${DATA_DIR}/archive"

# Directory used to store databases.
export DB_DIR="${DATA_DIR}/db"

# Directory used to store full-text indexed data.
export FTS_DIR="${DATA_DIR}/fts"
export FTS_EMAIL_DIR="${FTS_DIR}/email"

# Binary programs.
export BIN_DIR="/usr/local/bin"
export CMD_SPIDERD="${BIN_DIR}/spiderd"
export CMD_IMPORT="${BIN_DIR}/spider-import"

# Log directory.
export LOG_DIR="${DATA_DIR}/log"

# Secret key file used to encrypt/decrypt email messages and attachments.
export SECRET_KEY_FILE="${DATA_DIR}/secret.key"

# Service name.
export SRV="spiderd"

export OS="$(uname -s)"

# Install / Upgrade without confirm.
export WITHOUT_CONFIRM=NO

LOG() {
    echo -e "$@"
}

welcome() {
    cat <<EOF
**************************************************************
* Install or upgrade the ${PROG}
**************************************************************
We are going to check and create:

- Main config file: ${CONF}
- Secret key file: ${SECRET_KEY_FILE}
- System user and group: "${USER}", used to run Spider.
- ${DATA_DIR}/: root directory of application data.
- ${STORAGE_DIR}/: archived email messages and attachments.
- ${FTS_DIR}/: full-text indexed data.
- ${LOG_DIR}/: application log.

Existing config file and secret key will be preserved.

EOF

    if [[ ${WITHOUT_CONFIRM} != "YES" ]]; then
        echo -n "Press letter 'y' to continue, any other key to abort. "
        read GO
        if [[ ${GO} != "y" ]]; then
            LOG "Abort."
            exit 255
        fi
    fi
}

add_sys_user_group() {
    # Create daemon group.
    if ! grep "^${GROUP}:" /etc/group &>/dev/null; then
        if [[ "${OS}" == "Linux" ]]; then
            groupadd -g ${GROUP_ID} ${GROUP}
        elif [[ "${OS}" == "FreeBSD" ]]; then
            pw groupadd -g ${GROUP_ID} -n ${GROUP}
        elif [[ "${OS}" == "OpenBSD" ]]; then
            groupadd -g ${GROUP_ID} ${GROUP}
        fi
    fi

    # Create daemon user.
    if ! id ${USER} &>/dev/null; then
        if [[ "${OS}" == "Linux" ]]; then
            useradd \
                --no-create-home \
                -s /usr/sbin/nologin \
                -g ${GROUP} \
                -d ${DATA_DIR} \
                -u ${USER_ID} \
                ${USER}
        elif [[ "${OS}" == "FreeBSD" ]]; then
            pw useradd \
                -g ${GROUP} \
                -d ${DATA_DIR} \
                -u ${USER_ID} \
                -s /sbin/nologin \
                -n ${USER}
        elif [[ "${OS}" == "OpenBSD" ]]; then
            useradd \
                -u ${USER_ID} \
                -g ${GROUP} \
                -d ${DATA_DIR} \
                -s /sbin/nologin \
                ${USER}
        fi
    fi
}

create_conf_and_dirs() {
    [[ -f ${CONF} ]] &&
        chown ${USER}:${GROUP} ${CONF} && \
        chmod 0700 ${CONF}

    # Create directories and set correct owner/group and permission.
    for d in ${DATA_DIR} \
        ${DB_DIR} \
        ${FTS_DIR} \
        ${FTS_EMAIL_DIR} \
        ${STORAGE_DIR} \
        ${LOG_DIR}; do

        # Create if missing.
        # Could be a directory or symbol link.
        if [[ -d ${d} ]] || [[ -L ${d} ]]; then
            :
        else
            mkdir -p ${d}
        fi

        # Always set correct owner/group/permission.
        chown ${USER}:${GROUP} ${d}
        chmod 0700 ${d}
    done
}

handle_secret_key() {
    if [[ -e ${SECRET_KEY_FILE} ]]; then
        LOG "- Secret key file ${SECRET_KEY_FILE} exists. Not modified."
    else
        LOG "- Generate secret key file ${SECRET_KEY_FILE}, used to "
        LOG "  encrypt and decrypt email messages and attachments."

        if [[ "${OS}" == "Linux" ]]; then
            key="$(LC_CTYPE=C tr -dc [:alnum:] </dev/urandom | head -c 64)"
        elif [[ "${OS}" == "FreeBSD" ]]; then
            key="$(openssl rand -base64 64)"
        elif [[ "${OS}" == "OpenBSD" ]]; then
            key="$(LC_CTYPE=C </dev/random tr -cd [:alnum:] | fold -w 32 | head -1)"
        fi

        echo "${key}" > ${SECRET_KEY_FILE}
    fi

    # Always set owner/group/permission.
    chown ${USER}:${GROUP} ${SECRET_KEY_FILE}
    chmod 0400 ${SECRET_KEY_FILE}
}

copy_binary() {
    cp -f bin/spider* ${BIN_DIR}
    chown ${USER}:${GROUP} ${CMD_SPIDERD} ${CMD_IMPORT}
    chmod 0700 ${CMD_SPIDERD} ${CMD_IMPORT}
}

log_file() {
    # Create symbol link to log directory.
    [[ -e /var/log/spider ]] || ln -sf ${LOG_DIR} /var/log/spider
}

set_owner_group_mode() {
    # Make sure data has correct owner/group/permission.
    chown -R ${USER}:${GROUP} ${FTS_EMAIL_DIR}
    chmod 0700 ${FTS_EMAIL_DIR}

    # database and log directories
    chown -R ${USER}:${GROUP} ${DB_DIR} ${LOG_DIR} 2>/dev/null
    chmod -R 0700 ${DB_DIR} ${LOG_DIR} 2>/dev/null
}

enable_service() {
    if [[ "${OS}" == "Linux" ]]; then
        # Currently only systemd-based linux distros are supported.
        if [[ -d /lib/systemd/system ]]; then
            f="/lib/systemd/system/${SRV}.service"

            cp -f samples/rc/systemd.service ${f}
            chmod 0444 ${f}

            systemctl daemon-reload
            systemctl enable ${SRV}
        fi
    elif [[ "${OS}" == "FreeBSD" ]]; then
        [[ -d /usr/local/etc/rc.d/ ]] || mkdir -p /usr/local/etc/rc.d
        cp -f samples/rc/freebsd /usr/local/etc/rc.d/${SRV}
        sysrc -f /etc/rc.conf.local \
            ${SRV}_enable=YES \
            ${SRV}_user="${USER}" \
            ${SRV}_group="${GROUP}"
    elif [[ "${OS}" == "OpenBSD" ]]; then
        cp -f samples/rc/openbsd /etc/rc.d/${SRV}
        rcctl enable ${SRV}
    fi
}

restart_service() {
    if [[ "${OS}" == "Linux" ]]; then
        # Currently only systemd-based linux distros are supported.
        if [[ -d /lib/systemd/system ]]; then
            LOG "- Starting / Restarting the ${PROG}."
            systemctl restart ${SRV}
            sleep 2
            systemctl status ${SRV}

            if [[ $? != 0 ]]; then
                echo
                echo "**************************************************************"
                echo "********************     ERROR       *************************"
                echo "**************************************************************"
                echo "Failed to start the service, please check console output"
                echo "and log file (/var/log/messages or /var/log/syslog), contact"
                echo "support if necessary: https://spiderd.io/"

                exit 255
            fi
        fi
    elif [[ "${OS}" == "FreeBSD" ]]; then
        service ${SRV} restart
    elif [[ "${OS}" == "OpenBSD" ]]; then
        cp -f samples/rc/openbsd /etc/rc.d/${SRV}
        rcctl stop ${SRV}
        sleep 2
        rcctl start ${SRV}
    fi
}

finished() {
    LOG "- ${PROG} has been successfully installed / upgraded."
    LOG "- You can now access the web console at http://localhost:8080/"
}

for i in $@; do
    if [[ $i == "-y" ]]; then
        WITHOUT_CONFIRM=YES
        break
    fi
done

welcome
add_sys_user_group
create_conf_and_dirs
handle_secret_key
copy_binary
log_file
enable_service
set_owner_group_mode
restart_service
finished
